import 'dart:math' as math;

import 'package:knt_bloc/knt_bloc.dart';

import '../model/book.dart';
import '../repo/book_repo.dart';

class BookBloc extends BasePagingBloc<void, Book> {
  BookBloc() {
    watch<ChangeModeEvent>(
      _handleChangeModeEvent,
      transformer: restartable(),
    );
  }

  final _books = <Book>[];
  bool _usingIsolator = true;

  @override
  bool get needLog => true;

  @override
  BaseBackendInitializer<void, BookBlocBackend> launchBackend() =>
      (BackendArgument<void> argument) {
        return BookBlocBackend(argument);
      };

  @override
  Future<BaseResponsePaging<Book>> pagingInProgress(
      BaseRequestPaging<void> pagingQuery) async {
    await _loadBook();
    if (pagingQuery.offset >= _books.length) {
      return const BaseResponsePaging<Book>(dataList: []);
    }
    List<Book> partialBooks = _books.sublist(
      pagingQuery.offset,
      math.min(pagingQuery.offset + pagingQuery.limit, _books.length),
    );
    await Future.delayed(const Duration(milliseconds: 200));
    return BaseResponsePaging<Book>(dataList: partialBooks);
  }

  @override
  BasePagingResultState<Book> pagingStateInstance(
    List<Book> dataList,
    int offset, {
    String? tag,
  }) =>
      ShowBooksState(dataList, offset, tag: tag);

  @override
  CompareFunction<Book>? get comparable =>
      (b1, b2) => b1.title.compareTo(b2.title);

  Future<void> _handleChangeModeEvent(
    ChangeModeEvent event,
    Emitter<BaseState> emit,
  ) async {
    _usingIsolator = event.usingIsolate;
  }

  Future<void> _loadBook() async {
    if (_usingIsolator) {
      await initializeBackend();
      final Maybe<List<Book>> maybe = await run(event: 'LoadBooksEvent');
      if (maybe.hasValue) {
        _showBook(maybe.value);
      }
      if (maybe.hasError && maybe.error is KntException) {
        throw maybe.error;
      }
    } else {
      final loadedBook = await BookRepo().fetchBooks();
      _showBook(loadedBook);
    }
  }

  Future<void> _showBook(List<Book> books) async {
    _books.clear();
    _books.addAll(books);
  }
}

class BookBlocBackend extends BasePagingBlocBackend<Book> {
  BookBlocBackend(BackendArgument<void> argument) : super(argument);

  final _repo = BookRepo();

  @override
  void initActions() {
    super.initActions();
    whenEventCome('LoadBooksEvent').run(_loadBooks);
  }

  Future<List<Book>> _loadBooks({
    required String event,
    void data,
  }) =>
      _repo.fetchBooks();
}

// region EVENT
class ChangeModeEvent extends BaseEvent {
  const ChangeModeEvent({
    this.usingIsolate = false,
  });

  final bool usingIsolate;
}
// endregion

// region STATE
class ShowBooksState extends BasePagingResultState<Book> {
  const ShowBooksState(
    List<Book> dataList,
    int currentOffset, {
    String? tag,
  }) : super(dataList, currentOffset, tag: tag);
}
// endregion
